# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from taskflow.patterns import linear_flow
from taskflow.patterns import unordered_flow

from octavia.common import constants
from octavia.controller.worker.v2.tasks import amphora_driver_tasks
from octavia.controller.worker.v2.tasks import database_tasks
from octavia.controller.worker.v2.tasks import lifecycle_tasks
from octavia.controller.worker.v2.tasks import network_tasks


class ListenerFlows:

    def get_create_listener_flow(self, flavor_dict=None):
        """Create a flow to create a listener

        :returns: The flow for creating a listener
        """
        create_listener_flow = linear_flow.Flow(constants.CREATE_LISTENER_FLOW)
        create_listener_flow.add(lifecycle_tasks.ListenersToErrorOnRevertTask(
            requires=constants.LISTENERS))
        create_listener_flow.add(amphora_driver_tasks.ListenersUpdate(
            requires=constants.LOADBALANCER_ID))
        create_listener_flow.add(network_tasks.UpdateVIP(
            requires=constants.LISTENERS))

        if flavor_dict and flavor_dict.get(constants.SRIOV_VIP, False):
            create_listener_flow.add(*self._get_firewall_rules_subflow(
                flavor_dict))

        create_listener_flow.add(database_tasks.
                                 MarkLBAndListenersActiveInDB(
                                     requires=(constants.LOADBALANCER_ID,
                                               constants.LISTENERS)))
        return create_listener_flow

    def get_create_all_listeners_flow(self, flavor_dict=None):
        """Create a flow to create all listeners

        :returns: The flow for creating all listeners
        """
        create_all_listeners_flow = linear_flow.Flow(
            constants.CREATE_LISTENERS_FLOW)
        create_all_listeners_flow.add(
            database_tasks.GetListenersFromLoadbalancer(
                requires=constants.LOADBALANCER,
                provides=constants.LISTENERS))
        create_all_listeners_flow.add(database_tasks.ReloadLoadBalancer(
            requires=constants.LOADBALANCER_ID,
            provides=constants.LOADBALANCER))
        create_all_listeners_flow.add(amphora_driver_tasks.ListenersUpdate(
            requires=constants.LOADBALANCER_ID))
        create_all_listeners_flow.add(network_tasks.UpdateVIP(
            requires=constants.LISTENERS))

        if flavor_dict and flavor_dict.get(constants.SRIOV_VIP, False):
            create_all_listeners_flow.add(*self._get_firewall_rules_subflow(
                flavor_dict))

        create_all_listeners_flow.add(
            database_tasks.MarkHealthMonitorsOnlineInDB(
                requires=constants.LOADBALANCER))
        return create_all_listeners_flow

    def get_delete_listener_flow(self, flavor_dict=None):
        """Create a flow to delete a listener

        :returns: The flow for deleting a listener
        """
        delete_listener_flow = linear_flow.Flow(constants.DELETE_LISTENER_FLOW)
        delete_listener_flow.add(lifecycle_tasks.ListenerToErrorOnRevertTask(
            requires=constants.LISTENER))
        delete_listener_flow.add(amphora_driver_tasks.ListenerDelete(
            requires=constants.LISTENER))
        delete_listener_flow.add(network_tasks.UpdateVIPForDelete(
            requires=constants.LOADBALANCER_ID))
        delete_listener_flow.add(database_tasks.DeleteListenerInDB(
            requires=constants.LISTENER))

        if flavor_dict and flavor_dict.get(constants.SRIOV_VIP, False):
            delete_listener_flow.add(*self._get_firewall_rules_subflow(
                flavor_dict))

        delete_listener_flow.add(database_tasks.DecrementListenerQuota(
            requires=constants.PROJECT_ID))
        delete_listener_flow.add(database_tasks.MarkLBActiveInDBByListener(
            requires=constants.LISTENER))

        return delete_listener_flow

    def get_delete_listener_internal_flow(self, listener, flavor_dict=None):
        """Create a flow to delete a listener and l7policies internally

           (will skip deletion on the amp and marking LB active)

        :returns: The flow for deleting a listener
        """
        listener_id = listener[constants.LISTENER_ID]
        delete_listener_flow = linear_flow.Flow(
            constants.DELETE_LISTENER_FLOW + '-' + listener_id)
        # Should cascade delete all L7 policies
        delete_listener_flow.add(network_tasks.UpdateVIPForDelete(
            name='delete_update_vip_' + listener_id,
            requires=constants.LOADBALANCER_ID))
        delete_listener_flow.add(database_tasks.DeleteListenerInDB(
            name='delete_listener_in_db_' + listener_id,
            requires=constants.LISTENER,
            inject={constants.LISTENER: listener}))

        # Currently the flavor_dict will always be None since there is
        # no point updating the firewall rules when deleting the LB.
        # However, this may be used for additional flows in the future, so
        # adding this code for completeness.
        if flavor_dict and flavor_dict.get(constants.SRIOV_VIP, False):
            delete_listener_flow.add(*self._get_firewall_rules_subflow(
                flavor_dict))

        delete_listener_flow.add(database_tasks.DecrementListenerQuota(
            name='decrement_listener_quota_' + listener_id,
            requires=constants.PROJECT_ID))

        return delete_listener_flow

    def get_update_listener_flow(self, flavor_dict=None):
        """Create a flow to update a listener

        :returns: The flow for updating a listener
        """
        update_listener_flow = linear_flow.Flow(constants.UPDATE_LISTENER_FLOW)
        update_listener_flow.add(lifecycle_tasks.ListenerToErrorOnRevertTask(
            requires=constants.LISTENER))
        update_listener_flow.add(amphora_driver_tasks.ListenersUpdate(
            requires=constants.LOADBALANCER_ID))
        update_listener_flow.add(network_tasks.UpdateVIP(
            requires=constants.LISTENERS))

        if flavor_dict and flavor_dict.get(constants.SRIOV_VIP, False):
            update_listener_flow.add(*self._get_firewall_rules_subflow(
                flavor_dict))

        update_listener_flow.add(database_tasks.UpdateListenerInDB(
            requires=[constants.LISTENER, constants.UPDATE_DICT]))
        update_listener_flow.add(database_tasks.
                                 MarkLBAndListenersActiveInDB(
                                     requires=(constants.LOADBALANCER_ID,
                                               constants.LISTENERS)))

        return update_listener_flow

    def _get_firewall_rules_subflow(self, flavor_dict, timeout_dict=None):
        """Creates a subflow that updates the firewall rules in the amphorae.

        :returns: The subflow for updating firewall rules in the amphorae.
        """
        sf_name = constants.FIREWALL_RULES_SUBFLOW
        fw_rules_subflow = linear_flow.Flow(sf_name)

        fw_rules_subflow.add(database_tasks.GetAmphoraeFromLoadbalancer(
            name=sf_name + '-' + constants.GET_AMPHORAE_FROM_LB,
            requires=constants.LOADBALANCER_ID,
            provides=constants.AMPHORAE))

        fw_rules_subflow.add(
            amphora_driver_tasks.AmphoraeGetConnectivityStatus(
                name=(sf_name + '-' +
                      constants.AMPHORAE_GET_CONNECTIVITY_STATUS),
                requires=constants.AMPHORAE,
                inject={constants.TIMEOUT_DICT: timeout_dict,
                        constants.NEW_AMPHORA_ID: constants.NIL_UUID},
                provides=constants.AMPHORAE_STATUS))

        fw_rules_subflow.add(network_tasks.GetAmphoraeNetworkConfigs(
            name=sf_name + '-' + constants.GET_AMP_NETWORK_CONFIG,
            requires=constants.LOADBALANCER_ID,
            provides=constants.AMPHORAE_NETWORK_CONFIG))

        update_amps_subflow = unordered_flow.Flow(
            constants.AMP_UPDATE_FW_SUBFLOW)

        amp_0_subflow = linear_flow.Flow('amp-0-fw-update')

        amp_0_subflow.add(database_tasks.GetAmphoraFirewallRules(
            name=sf_name + '-0-' + constants.GET_AMPHORA_FIREWALL_RULES,
            requires=(constants.AMPHORAE, constants.AMPHORAE_NETWORK_CONFIG),
            provides=constants.AMPHORA_FIREWALL_RULES,
            inject={constants.AMPHORA_INDEX: 0}))

        amp_0_subflow.add(amphora_driver_tasks.SetAmphoraFirewallRules(
            name=sf_name + '-0-' + constants.SET_AMPHORA_FIREWALL_RULES,
            requires=(constants.AMPHORAE, constants.AMPHORA_FIREWALL_RULES,
                      constants.AMPHORAE_STATUS),
            inject={constants.AMPHORA_INDEX: 0,
                    constants.TIMEOUT_DICT: timeout_dict}))

        update_amps_subflow.add(amp_0_subflow)

        if (flavor_dict[constants.LOADBALANCER_TOPOLOGY] ==
                constants.TOPOLOGY_ACTIVE_STANDBY):

            amp_1_subflow = linear_flow.Flow('amp-1-fw-update')

            amp_1_subflow.add(database_tasks.GetAmphoraFirewallRules(
                name=sf_name + '-1-' + constants.GET_AMPHORA_FIREWALL_RULES,
                requires=(constants.AMPHORAE,
                          constants.AMPHORAE_NETWORK_CONFIG),
                provides=constants.AMPHORA_FIREWALL_RULES,
                inject={constants.AMPHORA_INDEX: 1}))

            amp_1_subflow.add(amphora_driver_tasks.SetAmphoraFirewallRules(
                name=sf_name + '-1-' + constants.SET_AMPHORA_FIREWALL_RULES,
                requires=(constants.AMPHORAE,
                          constants.AMPHORA_FIREWALL_RULES,
                          constants.AMPHORAE_STATUS),
                inject={constants.AMPHORA_INDEX: 1,
                        constants.TIMEOUT_DICT: timeout_dict}))

            update_amps_subflow.add(amp_1_subflow)

        fw_rules_subflow.add(update_amps_subflow)

        return fw_rules_subflow



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
import concurrent.futures
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from taskflow import engines as tf_engines
from taskflow.jobs.base import Job

from octavia.common import base_taskflow
import octavia.tests.unit.base as base


MAX_WORKERS = 1
ENGINE = 'parallel'

_engine_mock = mock.MagicMock()


class TestBaseTaskFlowEngine(base.TestCase):

    def setUp(self):

        conf = oslo_fixture.Config(cfg.CONF)
        conf.config(group="task_flow", max_workers=MAX_WORKERS)
        conf.config(group="task_flow", engine=ENGINE)
        conf.config(group="task_flow", disable_revert=True)
        super().setUp()

    @mock.patch('concurrent.futures.ThreadPoolExecutor',
                return_value='TESTEXECUTOR')
    @mock.patch('taskflow.engines.load',
                return_value=_engine_mock)
    def test_taskflow_load(self,
                           mock_tf_engine_load,
                           mock_ThreadPoolExecutor):

        # Test __init__

        base_taskflow_engine = base_taskflow.BaseTaskFlowEngine()

        concurrent.futures.ThreadPoolExecutor.assert_called_once_with(
            max_workers=MAX_WORKERS)

        # Test taskflow_load

        base_taskflow_engine.taskflow_load('TEST')

        tf_engines.load.assert_called_once_with(
            'TEST',
            engine=ENGINE,
            executor='TESTEXECUTOR',
            never_resolve=True)

        _engine_mock.compile.assert_called_once_with()
        _engine_mock.prepare.assert_called_once_with()


class TestTaskFlowServiceController(base.TestCase):

    _mock_uuid = '9a2ebc48-cd3e-429e-aa04-e32f5fc5442a'

    def setUp(self):
        self.conf = oslo_fixture.Config(cfg.CONF)
        self.conf.config(group="task_flow", engine='parallel')
        self.conf.config(group="task_flow", max_workers=MAX_WORKERS)
        self.driver_mock = mock.MagicMock()
        self.persistence_mock = mock.MagicMock()
        self.jobboard_mock = mock.MagicMock()
        self.driver_mock.job_board.return_value = self.jobboard_mock
        self.driver_mock.persistence_driver.get_persistence.return_value = (
            self.persistence_mock)
        self.service_controller = base_taskflow.TaskFlowServiceController(
            self.driver_mock)
        super().setUp()

    @mock.patch('oslo_utils.uuidutils.generate_uuid', return_value=_mock_uuid)
    @mock.patch('taskflow.engines.save_factory_details')
    def test_run_poster(self, mock_engines, mockuuid):
        flow_factory = mock.MagicMock()
        flow_factory.__name__ = 'testname'
        job_name = f'testname-{self._mock_uuid}'
        job_details = {'store': 'test'}
        with mock.patch.object(self.service_controller, '_wait_for_job'
                               ) as wait:
            uuid = self.service_controller.run_poster(flow_factory,
                                                      **job_details)
            save_logbook = self.persistence_mock.__enter__().get_connection(
            ).save_logbook
            save_logbook.assert_called()
            self.assertEqual(job_name, save_logbook.call_args[0][0].name)

            mock_engines.assert_called()
            save_args = mock_engines.call_args
            self.assertEqual(job_name, save_args[0][0].name)
            self.assertEqual(self._mock_uuid, save_args[0][0].uuid)
            self.assertEqual(flow_factory, save_args[0][1])
            self.assertEqual(self.persistence_mock.__enter__(),
                             save_args[1]['backend'])

            self.jobboard_mock.__enter__().post.assert_called()
            post_args = self.jobboard_mock.__enter__().post.call_args
            self.assertEqual(job_name, post_args[0][0])
            self.assertEqual(job_details, post_args[1]['details'])
            wait.assert_called()
            self.assertEqual(self._mock_uuid, uuid)

    def test__wait_for_job(self):
        job1 = mock.MagicMock()
        job2 = mock.MagicMock()
        job_board = mock.MagicMock()
        job_board.iterjobs.side_effect = [
            [job1, job2]
        ]
        self.service_controller._wait_for_job(job_board)

        job1.wait.assert_called_once()
        job2.wait.assert_called_once()

    @mock.patch('octavia.common.base_taskflow.'
                'ExtendExpiryDynamicLoggingConductor')
    @mock.patch('octavia.common.base_taskflow.DynamicLoggingConductor')
    @mock.patch('concurrent.futures.ThreadPoolExecutor')
    def test_run_conductor(self, mock_threadpoolexec, dynamiccond, expirycond):
        self.service_controller.run_conductor("test")
        expirycond.assert_called_once_with(
            "test", self.jobboard_mock.__enter__(),
            persistence=self.persistence_mock.__enter__(),
            engine='parallel',
            engine_options={
                'max_workers': MAX_WORKERS,
            })
        self.conf.config(group="task_flow",
                         jobboard_backend_driver='zookeeper_taskflow_driver')

        self.service_controller.run_conductor("test2")
        dynamiccond.assert_called_once_with(
            "test2", self.jobboard_mock.__enter__(),
            persistence=self.persistence_mock.__enter__(),
            engine='parallel')

    def test__extend_jobs(self):
        conductor = mock.MagicMock()
        conductor._name = 'mycontroller'

        job1 = mock.MagicMock()
        job1.expires_in.return_value = 10
        job2 = mock.MagicMock()
        job2.expires_in.return_value = 10
        job3 = mock.MagicMock()
        job3.expires_in.return_value = 30
        self.jobboard_mock.__enter__().iterjobs.return_value = [
            job1, job2, job3]

        self.jobboard_mock.__enter__().find_owner.side_effect = [
            'mycontroller',
            TypeError('no owner'),
            'mycontroller']

        self.service_controller._extend_jobs(conductor, 30)

        job1.extend_expiry.assert_called_once_with(30)
        job2.extend_expiry.assert_not_called()
        job3.extend_expiry.assert_not_called()


class TestJobDetailsFilter(base.TestCase):
    def test_filter(self):
        log_filter = base_taskflow.JobDetailsFilter()

        tls_container_data = {
            'certificate': '<CERTIFICATE>',
            'private_key': '<PRIVATE_KEY>',
            'passphrase': '<PASSPHRASE>',
            'intermediates': [
                '<INTERMEDIATE1>',
                '<INTERMEDIATE2>'
            ]
        }

        job = mock.Mock(spec=Job)
        job.details = {
            'store': {
                'listeners': [
                    {
                        'name': 'listener_name',
                        'default_tls_container_data': tls_container_data
                    }
                ],
                'any_recursive': {
                    'type': [
                        {
                            'other_list': [
                                tls_container_data,
                                {
                                    'test': tls_container_data,
                                }
                            ]
                        }
                    ]
                }
            }
        }

        record = mock.Mock()
        record.args = (job, 'something')

        ret = log_filter.filter(record)
        self.assertTrue(ret)

        self.assertNotIn(tls_container_data['certificate'], record.args[0])
        self.assertNotIn(tls_container_data['private_key'], record.args[0])
        self.assertNotIn(tls_container_data['passphrase'], record.args[0])
        self.assertNotIn(tls_container_data['intermediates'][0],
                         record.args[0])
        self.assertNotIn(tls_container_data['intermediates'][1],
                         record.args[0])
        self.assertIn('listener_name', record.args[0])

        record.args = ('arg1', 2)

        ret = log_filter.filter(record)
        self.assertTrue(ret)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from oslo_utils import uuidutils

from octavia.common import constants
from octavia.db import models
from octavia.network import data_models as network_models
from octavia.network.drivers.noop_driver import driver
import octavia.tests.unit.base as base


class TestNoopNetworkDriver(base.TestCase):
    FAKE_UUID_1 = uuidutils.generate_uuid()
    FAKE_UUID_2 = uuidutils.generate_uuid()
    FAKE_UUID_3 = uuidutils.generate_uuid()
    FAKE_UUID_4 = uuidutils.generate_uuid()
    FAKE_UUID_5 = uuidutils.generate_uuid()
    FAKE_UUID_6 = uuidutils.generate_uuid()
    FAKE_UUID_7 = uuidutils.generate_uuid()

    def setUp(self):
        super().setUp()
        self.mock_engine = mock.MagicMock()
        with mock.patch('octavia.network.drivers.noop_driver.driver.'
                        'create_engine') as mock_create_engine:
            mock_create_engine.return_value = self.mock_engine
            with mock.patch('octavia.network.drivers.noop_driver.'
                            'driver.event'):
                self.driver = driver.NoopNetworkDriver()
            mock_create_engine.assert_called_once_with(
                'sqlite:////tmp/octavia-network-noop.db')

        self.port = mock.MagicMock()
        self.port_id = 88
        self.port_name = 'port1'
        self.port.id = self.port_id
        self.network_id = self.FAKE_UUID_3
        self.network_name = 'net1'
        self.device_id = self.FAKE_UUID_4
        self.ip_address = "192.0.2.2"
        self.load_balancer = models.LoadBalancer()
        self.load_balancer.id = self.FAKE_UUID_2

        self.vip = models.Vip()
        self.vip.ip_address = "192.0.2.1"
        self.vip.subnet_id = uuidutils.generate_uuid()
        self.vip.port_id = uuidutils.generate_uuid()
        self.amphora_id = self.FAKE_UUID_1
        self.compute_id = self.FAKE_UUID_2
        self.compute2_id = self.FAKE_UUID_2
        self.subnet_id = self.FAKE_UUID_3
        self.subnet_name = 'subnet1'
        self.qos_policy_id = self.FAKE_UUID_5
        self.vrrp_port_id = self.FAKE_UUID_6

        self.amphora1 = models.Amphora()
        self.amphora1.id = uuidutils.generate_uuid()
        self.amphora1.compute_id = self.compute_id
        self.amphora1.vrrp_port_id = uuidutils.generate_uuid()
        self.amphora1.ha_port_id = uuidutils.generate_uuid()
        self.amphora1.vrrp_ip = '192.0.2.10'
        self.amphora1.ha_ip = '192.0.2.11'
        self.amphora2 = models.Amphora()
        self.amphora2.id = uuidutils.generate_uuid()
        self.amphora2.compute_id = self.compute2_id
        self.amphora2.vrrp_port_id = uuidutils.generate_uuid()
        self.amphora2.ha_port_id = uuidutils.generate_uuid()
        self.amphora2.vrrp_ip = '192.0.2.20'
        self.amphora2.ha_ip = '192.0.2.21'
        self.load_balancer.amphorae = [self.amphora1, self.amphora2]
        self.load_balancer.vip = self.vip
        self.subnet = mock.MagicMock()
        self.subnet.id = self.subnet_id

    def test_allocate_vip(self):
        self.driver.allocate_vip(self.load_balancer)
        self.assertEqual(
            (self.load_balancer, 'allocate_vip'),
            self.driver.driver.networkconfigconfig[self.load_balancer.id])

    def test_deallocate_vip(self):
        self.driver.deallocate_vip(self.vip)
        self.assertEqual((self.vip,
                          'deallocate_vip'),
                         self.driver.driver.networkconfigconfig[
                             self.vip.ip_address])

    def test_update_vip_sg(self):
        self.driver.update_vip_sg(self.load_balancer, self.vip)
        self.assertEqual((self.load_balancer, self.vip,
                          'update_vip_sg'),
                         self.driver.driver.networkconfigconfig[(
                             self.load_balancer.id, self.vip.ip_address)])

    def test_update_aap_port_sg(self):
        self.driver.update_aap_port_sg(self.load_balancer, self.amphora1,
                                       self.vip)
        self.assertEqual((self.load_balancer, self.vip, self.amphora1,
                          'update_aap_port_sg'),
                         self.driver.driver.networkconfigconfig[(
                             self.amphora1.id, self.vip.ip_address)])

    def test_unplug_vip(self):
        self.driver.unplug_vip(self.load_balancer, self.vip)
        self.assertEqual((self.load_balancer, self.vip,
                          'unplug_vip'),
                         self.driver.driver.networkconfigconfig[(
                             self.load_balancer.id, self.vip.ip_address)])

    def test_unplug_network(self):
        self.driver.unplug_network(self.compute_id, self.network_id)
        self.assertEqual((self.compute_id, self.network_id, 'unplug_network'),
                         self.driver.driver.networkconfigconfig[(
                             self.compute_id, self.network_id)])

    def test_get_plugged_networks(self):
        interface_db_mock = mock.MagicMock()
        interface_db_mock.port_id = self.port_id
        interface_db_mock.network_id = self.network_id
        interface_db_mock.compute_id = self.compute_id
        interface_db_mock.vnic_type = constants.VNIC_TYPE_NORMAL

        fixed_ips_db_mock = mock.MagicMock()
        fixed_ips_db_mock.port_id = self.port_id
        fixed_ips_db_mock.subnet_id = self.subnet_id
        fixed_ips_db_mock.ip_address = self.ip_address

        # mock out the sqlite db calls
        connect_mock = mock.MagicMock()
        connection_mock = mock.MagicMock()
        self.mock_engine.connect.return_value = connect_mock
        connect_mock.__enter__.return_value = connection_mock

        connection_mock.execute.side_effect = [[interface_db_mock],
                                               [fixed_ips_db_mock]]

        result = self.driver.get_plugged_networks(self.compute_id)

        self.assertEqual((self.compute_id, 'get_plugged_networks'),
                         self.driver.driver.networkconfigconfig[(
                             self.compute_id)])

        expected_fixed_ips = [network_models.FixedIP(
            subnet_id=self.subnet_id, ip_address=self.ip_address)]
        expected_interfaces = [network_models.Interface(
            compute_id=self.compute_id, network_id=self.network_id,
            port_id=self.port_id, fixed_ips=expected_fixed_ips,
            vnic_type=constants.VNIC_TYPE_NORMAL)]

        self.assertEqual(expected_interfaces, result)

    def test_update_vip(self):
        self.driver.update_vip(self.load_balancer)
        self.assertEqual((self.load_balancer, False, 'update_vip'),
                         self.driver.driver.networkconfigconfig[(
                             self.load_balancer.id
                         )])

    def test_get_network(self):
        network = self.driver.get_network(self.network_id)
        self.assertEqual(
            (self.network_id, 'get_network'),
            self.driver.driver.networkconfigconfig[self.network_id]
        )
        self.assertEqual(self.network_id, network.id)
        network_again = self.driver.get_network(self.network_id)
        self.assertEqual(network, network_again)

    def test_get_subnet(self):
        subnet = self.driver.get_subnet(self.subnet_id)
        self.assertEqual(
            (self.subnet_id, 'get_subnet'),
            self.driver.driver.networkconfigconfig[self.subnet_id]
        )
        self.assertEqual(self.subnet_id, subnet.id)
        subnet_again = self.driver.get_subnet(self.subnet_id)
        self.assertEqual(subnet, subnet_again)

    def test_get_port(self):
        port = self.driver.get_port(self.port_id)
        self.assertEqual(
            (self.port_id, 'get_port'),
            self.driver.driver.networkconfigconfig[self.port_id]
        )
        self.assertEqual(self.port_id, port.id)
        port_again = self.driver.get_port(self.port_id)
        self.assertEqual(port, port_again)

    def test_get_network_by_name(self):
        network = self.driver.get_network_by_name(self.network_name)
        self.assertEqual(
            (self.network_name, 'get_network_by_name'),
            self.driver.driver.networkconfigconfig[self.network_name]
        )
        self.assertEqual(self.network_name, network.name)
        network_again = self.driver.get_network_by_name(self.network_name)
        self.assertEqual(network, network_again)

    def test_get_subnet_by_name(self):
        subnet = self.driver.get_subnet_by_name(self.subnet_name)
        self.assertEqual(
            (self.subnet_name, 'get_subnet_by_name'),
            self.driver.driver.networkconfigconfig[self.subnet_name]
        )
        self.assertEqual(self.subnet_name, subnet.name)
        subnet_again = self.driver.get_subnet_by_name(self.subnet_name)
        self.assertEqual(subnet, subnet_again)

    def test_get_port_by_name(self):
        port = self.driver.get_port_by_name(self.port_name)
        self.assertEqual(
            (self.port_name, 'get_port_by_name'),
            self.driver.driver.networkconfigconfig[self.port_name]
        )
        self.assertEqual(self.port_name, port.name)
        port_again = self.driver.get_port_by_name(self.port_name)
        self.assertEqual(port, port_again)

    def test_get_port_by_net_id_device_id(self):
        port = self.driver.get_port_by_net_id_device_id(
            self.network_id, self.device_id)
        self.assertEqual(
            (self.network_id, self.device_id,
             'get_port_by_net_id_device_id'),
            self.driver.driver.networkconfigconfig[(self.network_id,
                                                    self.device_id)]
        )
        self.assertEqual(self.network_id, port.network_id)
        self.assertEqual(self.device_id, port.device_id)
        port_again = self.driver.get_port_by_net_id_device_id(
            self.network_id, self.device_id)
        self.assertEqual(port, port_again)

    def test_get_security_group(self):
        FAKE_SG_NAME = 'fake_sg_name'
        result = self.driver.get_security_group(FAKE_SG_NAME)

        self.assertEqual((FAKE_SG_NAME, 'get_security_group'),
                         self.driver.driver.networkconfigconfig[FAKE_SG_NAME])
        self.assertTrue(uuidutils.is_uuid_like(result.id))

    def test_plug_port(self):
        self.driver.plug_port(self.amphora1, self.port)
        self.assertEqual(
            (self.amphora1, self.port, 'plug_port'),
            self.driver.driver.networkconfigconfig[self.amphora1.id,
                                                   self.port.id]
        )

    def test_get_network_configs(self):
        amp_config = self.driver.get_network_configs(self.load_balancer)
        self.assertEqual(
            (self.load_balancer, 'get_network_configs'),
            self.driver.driver.networkconfigconfig[self.load_balancer.id]
        )
        self.assertEqual(2, len(amp_config))
        self.assertEqual(self.amphora1, amp_config[self.amphora1.id].amphora)
        self.assertEqual(self.amphora2, amp_config[self.amphora2.id].amphora)

    def test_get_qos_policy(self):
        self.driver.get_qos_policy(self.qos_policy_id)
        self.assertEqual(
            (self.qos_policy_id, 'get_qos_policy'),
            self.driver.driver.networkconfigconfig[self.qos_policy_id]
        )

    def test_apply_qos_on_port(self):
        self.driver.apply_qos_on_port(self.qos_policy_id, self.vrrp_port_id)
        self.assertEqual(
            (self.qos_policy_id, self.vrrp_port_id, 'apply_qos_on_port'),
            self.driver.driver.networkconfigconfig[self.qos_policy_id,
                                                   self.vrrp_port_id]
        )

    def test_plug_aap_port(self):
        self.driver.plug_aap_port(self.load_balancer, self.vip, self.amphora1,
                                  self.subnet)
        self.assertEqual(
            (self.load_balancer, self.vip, self.amphora1, self.subnet,
             'plug_aap_port'),
            self.driver.driver.networkconfigconfig[self.amphora1.id,
                                                   self.vip.ip_address]
        )

    def test_unplug_aap(self):
        self.driver.unplug_aap_port(self.vip, self.amphora1, self.subnet)
        self.assertEqual(
            (self.vip, self.amphora1, self.subnet,
             'unplug_aap_port'),
            self.driver.driver.networkconfigconfig[self.amphora1.id,
                                                   self.vip.ip_address]
        )

    def test_delete_port(self):
        PORT_ID = uuidutils.generate_uuid()

        self.driver.delete_port(PORT_ID)

        self.assertEqual((PORT_ID, 'delete_port'),
                         self.driver.driver.networkconfigconfig[PORT_ID])

    def test_set_port_admin_state_up(self):
        PORT_ID = uuidutils.generate_uuid()

        self.driver.set_port_admin_state_up(PORT_ID, False)

        self.assertEqual(
            (PORT_ID, False, 'admin_down_port'),
            self.driver.driver.networkconfigconfig[(PORT_ID, False)])

    def test_create_port(self):
        FAKE_NAME = 'fake_name'
        IP_ADDRESS = '2001:db8::77'
        NETWORK_ID = uuidutils.generate_uuid()
        QOS_POLICY_ID = uuidutils.generate_uuid()
        SUBNET_ID = uuidutils.generate_uuid()
        FIXED_IPS = [{'ip_address': IP_ADDRESS, 'subnet_id': SUBNET_ID},
                     {'subnet_id': SUBNET_ID}]

        # Test minimum
        result = self.driver.create_port(NETWORK_ID)

        self.assertIsInstance(result, network_models.Port)
        self.assertEqual(NETWORK_ID, result.network_id)

        # Test full parameters
        result = self.driver.create_port(
            NETWORK_ID, name=FAKE_NAME, fixed_ips=FIXED_IPS,
            admin_state_up=False, qos_policy_id=QOS_POLICY_ID)

        self.assertIsInstance(result, network_models.Port)
        self.assertEqual(NETWORK_ID, result.network_id)
        self.assertEqual(FAKE_NAME, result.name)
        self.assertEqual(IP_ADDRESS, result.fixed_ips[0].ip_address)
        self.assertEqual(SUBNET_ID, result.fixed_ips[0].subnet_id)
        self.assertEqual('198.51.100.56', result.fixed_ips[1].ip_address)
        self.assertEqual(SUBNET_ID, result.fixed_ips[1].subnet_id)
        self.assertEqual(QOS_POLICY_ID, result.qos_policy_id)
        self.assertFalse(result.admin_state_up)

    def test_plug_fixed_ip(self):
        self.driver.plug_fixed_ip(self.port_id, self.subnet_id,
                                  self.ip_address)
        self.assertEqual(
            (self.port_id, self.subnet_id, self.ip_address, 'plug_fixed_ip'),
            self.driver.driver.networkconfigconfig[
                self.port_id, self.subnet_id]
        )



# Copyright 2024 NTT DATA Group Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from unittest import mock

from oslo_config import cfg

from octavia.controller.worker.v2 import taskflow_jobboard_driver
import octavia.tests.unit.base as base


class TestRedisTaskFlowDriver(base.TestCase):

    @mock.patch('octavia.controller.worker.v2.taskflow_jobboard_driver.'
                'job_backends')
    def test_job_board_default(self, mock_job_backends):
        driver = taskflow_jobboard_driver.RedisTaskFlowDriver(mock.Mock())
        driver.job_board(None)
        mock_job_backends.backend.assert_called_once_with(
            'octavia_jobboard',
            {
                'board': 'redis',
                'host': '127.0.0.1',
                'port': 6379,
                'db': 0,
                'namespace': 'octavia_jobboard',
                'sentinel': None,
                'sentinel_fallbacks': [],
                'ssl': False,
                'ssl_keyfile': None,
                'ssl_certfile': None,
                'ssl_ca_certs': None,
                'ssl_cert_reqs': 'required',
                'sentinel_kwargs': {
                    'ssl': False,
                    'ssl_keyfile': None,
                    'ssl_certfile': None,
                    'ssl_ca_certs': None,
                    'ssl_cert_reqs': 'required',
                }
            },
            persistence=None
        )

    @mock.patch('octavia.controller.worker.v2.taskflow_jobboard_driver.'
                'job_backends')
    def test_job_board_password(self, mock_job_backends):
        driver = taskflow_jobboard_driver.RedisTaskFlowDriver(mock.Mock())
        cfg.CONF.set_override('jobboard_backend_password', 'redispass',
                              group='task_flow')
        driver.job_board(None)
        mock_job_backends.backend.assert_called_once_with(
            'octavia_jobboard',
            {
                'board': 'redis',
                'host': '127.0.0.1',
                'port': 6379,
                'db': 0,
                'namespace': 'octavia_jobboard',
                'password': 'redispass',
                'sentinel': None,
                'sentinel_fallbacks': [],
                'ssl': False,
                'ssl_keyfile': None,
                'ssl_certfile': None,
                'ssl_ca_certs': None,
                'ssl_cert_reqs': 'required',
                'sentinel_kwargs': {
                    'ssl': False,
                    'ssl_keyfile': None,
                    'ssl_certfile': None,
                    'ssl_ca_certs': None,
                    'ssl_cert_reqs': 'required',
                }
            },
            persistence=None
        )

    @mock.patch('octavia.controller.worker.v2.taskflow_jobboard_driver.'
                'job_backends')
    def test_job_board_username(self, mock_job_backends):
        driver = taskflow_jobboard_driver.RedisTaskFlowDriver(mock.Mock())
        cfg.CONF.set_override('jobboard_backend_password', 'redispass',
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_username', 'redisuser',
                              group='task_flow')
        driver.job_board(None)
        mock_job_backends.backend.assert_called_once_with(
            'octavia_jobboard',
            {
                'board': 'redis',
                'host': '127.0.0.1',
                'port': 6379,
                'db': 0,
                'namespace': 'octavia_jobboard',
                'username': 'redisuser',
                'password': 'redispass',
                'sentinel': None,
                'sentinel_fallbacks': [],
                'ssl': False,
                'ssl_keyfile': None,
                'ssl_certfile': None,
                'ssl_ca_certs': None,
                'ssl_cert_reqs': 'required',
                'sentinel_kwargs': {
                    'ssl': False,
                    'ssl_keyfile': None,
                    'ssl_certfile': None,
                    'ssl_ca_certs': None,
                    'ssl_cert_reqs': 'required',
                }
            },
            persistence=None
        )

    @mock.patch('octavia.controller.worker.v2.taskflow_jobboard_driver.'
                'job_backends')
    def test_job_board_ssl(self, mock_job_backends):
        driver = taskflow_jobboard_driver.RedisTaskFlowDriver(mock.Mock())
        cfg.CONF.set_override(
            'jobboard_redis_backend_ssl_options',
            {
                'ssl': True,
                'ssl_keyfile': 'rediskey',
                'ssl_certfile': 'rediscert',
                'ssl_ca_certs': 'redisca',
                'ssl_cert_reqs': 'required'
            },
            group='task_flow')
        driver.job_board(None)
        mock_job_backends.backend.assert_called_once_with(
            'octavia_jobboard',
            {
                'board': 'redis',
                'host': '127.0.0.1',
                'port': 6379,
                'db': 0,
                'namespace': 'octavia_jobboard',
                'sentinel': None,
                'sentinel_fallbacks': [],
                'ssl': True,
                'ssl_keyfile': 'rediskey',
                'ssl_certfile': 'rediscert',
                'ssl_ca_certs': 'redisca',
                'ssl_cert_reqs': 'required',
                'sentinel_kwargs': {
                    'ssl': False,
                    'ssl_keyfile': None,
                    'ssl_certfile': None,
                    'ssl_ca_certs': None,
                    'ssl_cert_reqs': 'required',
                }
            },
            persistence=None
        )

    @mock.patch('octavia.controller.worker.v2.taskflow_jobboard_driver.'
                'job_backends')
    def test_job_board_sentinel(self, mock_job_backends):
        driver = taskflow_jobboard_driver.RedisTaskFlowDriver(mock.Mock())
        cfg.CONF.set_override('jobboard_redis_sentinel', 'mymaster',
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_hosts',
                              ['host1', 'host2', 'host3'],
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_port', 26379,
                              group='task_flow')
        driver.job_board(None)
        mock_job_backends.backend.assert_called_once_with(
            'octavia_jobboard',
            {
                'board': 'redis',
                'host': 'host1',
                'port': 26379,
                'db': 0,
                'namespace': 'octavia_jobboard',
                'sentinel': 'mymaster',
                'sentinel_fallbacks': ['host2:26379', 'host3:26379'],
                'ssl': False,
                'ssl_keyfile': None,
                'ssl_certfile': None,
                'ssl_ca_certs': None,
                'ssl_cert_reqs': 'required',
                'sentinel_kwargs': {
                    'ssl': False,
                    'ssl_keyfile': None,
                    'ssl_certfile': None,
                    'ssl_ca_certs': None,
                    'ssl_cert_reqs': 'required',
                }
            },
            persistence=None
        )

    @mock.patch('octavia.controller.worker.v2.taskflow_jobboard_driver.'
                'job_backends')
    def test_job_board_sentinel_password(self, mock_job_backends):
        driver = taskflow_jobboard_driver.RedisTaskFlowDriver(mock.Mock())
        cfg.CONF.set_override('jobboard_redis_sentinel', 'mymaster',
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_hosts',
                              ['host1', 'host2', 'host3'],
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_port', 26379,
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_password', 'redispass',
                              group='task_flow')
        cfg.CONF.set_override('jobboard_redis_sentinel_password',
                              'sentinelpass', group='task_flow')
        driver.job_board(None)
        mock_job_backends.backend.assert_called_once_with(
            'octavia_jobboard',
            {
                'board': 'redis',
                'host': 'host1',
                'port': 26379,
                'db': 0,
                'namespace': 'octavia_jobboard',
                'password': 'redispass',
                'sentinel': 'mymaster',
                'sentinel_fallbacks': ['host2:26379', 'host3:26379'],
                'ssl': False,
                'ssl_keyfile': None,
                'ssl_certfile': None,
                'ssl_ca_certs': None,
                'ssl_cert_reqs': 'required',
                'sentinel_kwargs': {
                    'password': 'sentinelpass',
                    'ssl': False,
                    'ssl_keyfile': None,
                    'ssl_certfile': None,
                    'ssl_ca_certs': None,
                    'ssl_cert_reqs': 'required',
                }
            },
            persistence=None
        )

    @mock.patch('octavia.controller.worker.v2.taskflow_jobboard_driver.'
                'job_backends')
    def test_job_board_sentinel_username(self, mock_job_backends):
        driver = taskflow_jobboard_driver.RedisTaskFlowDriver(mock.Mock())
        cfg.CONF.set_override('jobboard_redis_sentinel', 'mymaster',
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_hosts',
                              ['host1', 'host2', 'host3'],
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_port', 26379,
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_username', 'redisuser',
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_password', 'redispass',
                              group='task_flow')
        cfg.CONF.set_override('jobboard_redis_sentinel_username',
                              'sentineluser', group='task_flow')
        cfg.CONF.set_override('jobboard_redis_sentinel_password',
                              'sentinelpass', group='task_flow')
        driver.job_board(None)
        mock_job_backends.backend.assert_called_once_with(
            'octavia_jobboard',
            {
                'board': 'redis',
                'host': 'host1',
                'port': 26379,
                'db': 0,
                'namespace': 'octavia_jobboard',
                'username': 'redisuser',
                'password': 'redispass',
                'sentinel': 'mymaster',
                'sentinel_fallbacks': ['host2:26379', 'host3:26379'],
                'ssl': False,
                'ssl_keyfile': None,
                'ssl_certfile': None,
                'ssl_ca_certs': None,
                'ssl_cert_reqs': 'required',
                'sentinel_kwargs': {
                    'username': 'sentineluser',
                    'password': 'sentinelpass',
                    'ssl': False,
                    'ssl_keyfile': None,
                    'ssl_certfile': None,
                    'ssl_ca_certs': None,
                    'ssl_cert_reqs': 'required',
                }
            },
            persistence=None
        )

    @mock.patch('octavia.controller.worker.v2.taskflow_jobboard_driver.'
                'job_backends')
    def test_job_board_sentinel_ssl(self, mock_job_backends):
        driver = taskflow_jobboard_driver.RedisTaskFlowDriver(mock.Mock())
        cfg.CONF.set_override('jobboard_redis_sentinel', 'mymaster',
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_hosts',
                              ['host1', 'host2', 'host3'],
                              group='task_flow')
        cfg.CONF.set_override('jobboard_backend_port', 26379,
                              group='task_flow')
        cfg.CONF.set_override(
            'jobboard_redis_backend_ssl_options',
            {
                'ssl': True,
                'ssl_keyfile': 'rediskey',
                'ssl_certfile': 'rediscert',
                'ssl_ca_certs': 'redisca',
                'ssl_cert_reqs': 'required'
            },
            group='task_flow')
        cfg.CONF.set_override(
            'jobboard_redis_sentinel_ssl_options',
            {
                'ssl': True,
                'ssl_keyfile': 'sentinelkey',
                'ssl_certfile': 'sentinelcert',
                'ssl_ca_certs': 'sentinelca',
                'ssl_cert_reqs': 'required'
            },
            group='task_flow')
        driver.job_board(None)
        mock_job_backends.backend.assert_called_once_with(
            'octavia_jobboard',
            {
                'board': 'redis',
                'host': 'host1',
                'port': 26379,
                'db': 0,
                'namespace': 'octavia_jobboard',
                'sentinel': 'mymaster',
                'sentinel_fallbacks': ['host2:26379', 'host3:26379'],
                'ssl': True,
                'ssl_keyfile': 'rediskey',
                'ssl_certfile': 'rediscert',
                'ssl_ca_certs': 'redisca',
                'ssl_cert_reqs': 'required',
                'sentinel_kwargs': {
                    'ssl': True,
                    'ssl_keyfile': 'sentinelkey',
                    'ssl_certfile': 'sentinelcert',
                    'ssl_ca_certs': 'sentinelca',
                    'ssl_cert_reqs': 'required'
                }
            },
            persistence=None
        )



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from typing import Optional

from wsme import types as wtypes

from oslo_db.sqlalchemy import models
from oslo_utils import strutils
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy.orm import collections
from sqlalchemy.orm import declarative_base

from octavia.common import constants


class OctaviaBase(models.ModelBase):

    __data_model__ = None

    @staticmethod
    def _get_unique_key(obj):
        """Returns a unique key for passed object for data model building."""
        # First handle all objects with their own ID, then handle subordinate
        # objects.
        if obj.__class__.__name__ in ['Member', 'Pool', 'LoadBalancer',
                                      'Listener', 'Amphora', 'L7Policy',
                                      'L7Rule', 'Flavor', 'FlavorProfile',
                                      'AvailabilityZoneProfile']:
            return obj.__class__.__name__ + obj.id
        if obj.__class__.__name__ in ['SessionPersistence', 'HealthMonitor']:
            return obj.__class__.__name__ + obj.pool_id
        if obj.__class__.__name__ in ['ListenerStatistics']:
            return obj.__class__.__name__ + obj.listener_id + obj.amphora_id
        if obj.__class__.__name__ in ['ListenerCidr']:
            return obj.__class__.__name__ + obj.listener_id + obj.cidr
        if obj.__class__.__name__ in ['VRRPGroup', 'Vip']:
            return obj.__class__.__name__ + obj.load_balancer_id
        if obj.__class__.__name__ in ['AmphoraHealth']:
            return obj.__class__.__name__ + obj.amphora_id
        if obj.__class__.__name__ in ['SNI']:
            return (obj.__class__.__name__ +
                    obj.listener_id + obj.tls_container_id)
        if obj.__class__.__name__ in ['Quotas']:
            return obj.__class__.__name__ + obj.project_id
        if obj.__class__.__name__ in ['AvailabilityZone']:
            return obj.__class__.__name__ + obj.name
        if obj.__class__.__name__ in ['AdditionalVip']:
            return (obj.__class__.__name__ +
                    obj.load_balancer_id + obj.subnet_id)
        if obj.__class__.__name__ in ['VipSecurityGroup']:
            return obj.__class__.__name__ + obj.load_balancer_id + obj.sg_id
        if obj.__class__.__name__ in ['AmphoraMemberPort']:
            return obj.__class__.__name__ + obj.port_id
        raise NotImplementedError

    def to_data_model(
            self, _graph_nodes=None, recursion_depth: Optional[int] = None):
        """Converts to a data model graph.

        In order to make the resulting data model graph usable no matter how
        many internal references are followed, we generate a complete graph of
        OctaviaBase nodes connected to the object passed to this method.

        :param _graph_nodes: Used only for internal recursion of this
                             method. Should not be called from the outside.
                             Contains a dictionary of all OctaviaBase type
                             objects in the generated graph
        :param recursion_depth: Used only for configuring recursion.
                                This option allows to limit recursion depth.
                                It could be used when we need only main node
                                and its first level relationships.
                                It allows to save time on recursion calls for
                                huge graphs, when only main object is
                                necessary.
        """
        _graph_nodes = _graph_nodes or {}
        if not self.__data_model__:
            raise NotImplementedError
        dm_kwargs = {}
        for column in self.__table__.columns:
            dm_kwargs[column.name] = getattr(self, column.name)

        attr_names = [attr_name for attr_name in dir(self)
                      if not attr_name.startswith('_')]
        # Appending early, as any unique ID should be defined already and
        # the rest of this object will get filled out more fully later on,
        # and we need to add ourselves to the _graph_nodes before we
        # attempt recursion.
        dm_self = self.__data_model__(**dm_kwargs)
        dm_key = self._get_unique_key(dm_self)
        _graph_nodes.update({dm_key: dm_self})
        new_depth = recursion_depth
        need_recursion = recursion_depth is None or recursion_depth > 0
        # decrease depth of recursion on new recursion call
        if new_depth:
            new_depth -= 1
        for attr_name in attr_names:
            attr = getattr(self, attr_name)
            if (need_recursion and
                    isinstance(attr, OctaviaBase) and
                    attr.__class__):
                # If this attr is already in the graph node list, just
                # reference it there and don't recurse.
                ukey = self._get_unique_key(attr)
                if ukey in _graph_nodes.keys():
                    setattr(dm_self, attr_name, _graph_nodes[ukey])
                else:
                    setattr(dm_self, attr_name, attr.to_data_model(
                        _graph_nodes=_graph_nodes,
                        recursion_depth=new_depth))
            elif isinstance(attr, (collections.InstrumentedList, list)):
                setattr(dm_self, attr_name, [])
                listref = getattr(dm_self, attr_name)
                for item in attr:
                    if (need_recursion and
                            isinstance(item, OctaviaBase) and
                            item.__class__):
                        ukey = self._get_unique_key(item)
                        if ukey in _graph_nodes.keys():
                            listref.append(_graph_nodes[ukey])
                        else:
                            listref.append(
                                item.to_data_model(_graph_nodes=_graph_nodes,
                                                   recursion_depth=new_depth))
                    elif not isinstance(item, OctaviaBase):
                        listref.append(item)
        return dm_self

    @staticmethod
    def apply_filter(query, model, filters):
        # Convert boolean filters to proper type
        for key in filters:
            attr = getattr(model.__v2_wsme__, key, None)
            if isinstance(attr, wtypes.wsattr) and attr.datatype == bool:
                filters[key] = strutils.bool_from_string(filters[key])
        # Special case for 'enabled', it's 'admin_state_up' in the WSME class
        # definition and the attribute has already been renamed to 'enabled' by
        # a previous pagination filter
        if constants.ENABLED in filters:
            filters[constants.ENABLED] = strutils.bool_from_string(
                filters[constants.ENABLED])

        translated_filters = {}
        child_map = {}
        for attr, name_map in model.__v2_wsme__._child_map.items():
            for k, v in name_map.items():
                if attr in filters and k in filters[attr]:
                    child_map.setdefault(attr, {}).update(
                        {k: filters[attr].pop(k)})
            filters.pop(attr, None)

        for k, v in model.__v2_wsme__._type_to_model_map.items():
            if k in filters:
                translated_filters[v] = filters.pop(k)
        translated_filters.update(filters)
        if translated_filters:
            query = query.filter_by(**translated_filters)
        for k, v in child_map.items():
            query = query.join(getattr(model, k)).filter_by(**v)
        return query

    def __repr__(self):
        params = sorted(
            (k, getattr(self, k)) for k in self.__mapper__.columns.keys()
        )
        params = ", ".join(f"{k}={v!r}" for k, v in params)
        return f"{self.__class__.__name__}({params})"


class LookupTableMixin:
    """Mixin to add to classes that are lookup tables."""
    name = sa.Column(sa.String(255), primary_key=True, nullable=False)
    description = sa.Column(sa.String(255), nullable=True)


class IdMixin:
    """Id mixin, add to subclasses that have an id."""
    id = sa.Column(sa.String(36), primary_key=True,
                   default=uuidutils.generate_uuid)


class ProjectMixin:
    """Tenant mixin, add to subclasses that have a project."""
    project_id = sa.Column(sa.String(36))


class NameMixin:
    """Name mixin to add to classes which need a name."""
    name = sa.Column(sa.String(255), nullable=True)


class TagMixin:
    """Tags mixin to add to classes which need tags.

    The class must realize the specified db relationship as well.
    """

    @property
    def tags(self):
        if self._tags:
            return [each_tag.tag for each_tag in self._tags]
        return []

    @tags.setter
    def tags(self, values):
        new_tags = []
        if values:
            for tag in values:
                tag_ref = Tags()
                tag_ref.resource_id = self.id
                tag_ref.tag = tag
                new_tags.append(tag_ref)
        self._tags = new_tags


BASE = declarative_base(cls=OctaviaBase)


class Tags(BASE):
    __tablename__ = "tags"

    resource_id = sa.Column(sa.String(36), primary_key=True)
    tag = sa.Column(sa.String(255), primary_key=True, index=True)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright (c) 2015 Rackspace
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import functools
import hashlib
import json
import os
import ssl
import time
from typing import Optional
import warnings

from oslo_context import context as oslo_context
from oslo_log import log as logging
import requests
from stevedore import driver as stevedore_driver

from octavia.amphorae.driver_exceptions import exceptions as driver_except
from octavia.amphorae.drivers import driver_base
from octavia.amphorae.drivers.haproxy import exceptions as exc
from octavia.amphorae.drivers.keepalived import vrrp_rest_driver
from octavia.common.config import cfg
from octavia.common import constants as consts
import octavia.common.jinja.haproxy.combined_listeners.jinja_cfg as jinja_combo
from octavia.common.jinja.lvs import jinja_cfg as jinja_udp_cfg
from octavia.common.tls_utils import cert_parser
from octavia.common import utils
from octavia.db import api as db_api
from octavia.db import models as db_models
from octavia.db import repositories as repo


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class HaproxyAmphoraLoadBalancerDriver(
    driver_base.AmphoraLoadBalancerDriver,
        vrrp_rest_driver.KeepalivedAmphoraDriverMixin):

    def __init__(self):
        super().__init__()
        self.clients = {
            'base': AmphoraAPIClientBase(),
            '1.0': AmphoraAPIClient1_0(),
        }
        self.cert_manager = stevedore_driver.DriverManager(
            namespace='octavia.cert_manager',
            name=CONF.certificates.cert_manager,
            invoke_on_load=True,
        ).driver

        self.jinja_combo = jinja_combo.JinjaTemplater(
            base_amp_path=CONF.haproxy_amphora.base_path,
            base_crt_dir=CONF.haproxy_amphora.base_cert_dir,
            haproxy_template=CONF.haproxy_amphora.haproxy_template,
            connection_logging=CONF.haproxy_amphora.connection_logging)
        self.lvs_jinja = jinja_udp_cfg.LvsJinjaTemplater()

    def _get_haproxy_versions(self, amphora, timeout_dict=None):
        """Get major and minor version number from haproxy

        Example: ['1', '6']

        :returns version_list: A list with the major and minor numbers
        """
        self._populate_amphora_api_version(
            amphora, timeout_dict=timeout_dict)
        amp_info = self.clients[amphora.api_version].get_info(
            amphora, timeout_dict=timeout_dict)
        haproxy_version_string = amp_info['haproxy_version']

        return haproxy_version_string.split('.')[:2]

    def _populate_amphora_api_version(self, amphora, timeout_dict=None,
                                      raise_retry_exception=False):
        """Populate the amphora object with the api_version

        This will query the amphora for version discovery and populate
        the api_version string attribute on the amphora object.

        :returns: None
        """
        if not getattr(amphora, 'api_version', None):
            try:
                amphora.api_version = self.clients['base'].get_api_version(
                    amphora, timeout_dict=timeout_dict,
                    raise_retry_exception=raise_retry_exception)['api_version']
            except exc.NotFound:
                # Amphora is too old for version discovery, default to 0.5
                amphora.api_version = '0.5'
        LOG.debug('Amphora %s has API version %s',
                  amphora.id, amphora.api_version)
        api_version = list(map(int, amphora.api_version.split('.')))

        if api_version[0] == 0 and api_version[1] <= 5:  # 0.5 or earlier
            raise driver_except.AmpVersionUnsupported(
                version=amphora.api_version)

        return api_version

    def check(self, amphora: db_models.Amphora,
              timeout_dict: Optional[dict] = None):
        """Check connectivity to the amphora."""
        self._populate_amphora_api_version(amphora, timeout_dict)

    def update_amphora_listeners(self, loadbalancer, amphora,
                                 timeout_dict=None):
        """Update the amphora with a new configuration.

        :param loadbalancer: The load balancer to update
        :type loadbalancer: object
        :param amphora: The amphora to update
        :type amphora: object
        :param timeout_dict: Dictionary of timeout values for calls to the
                             amphora. May contain: req_conn_timeout,
                             req_read_timeout, conn_max_retries,
                             conn_retry_interval
        :returns: None

        Updates the configuration of the listeners on a single amphora.
        """
        # if the amphora does not yet have listeners, no need to update them.
        if not loadbalancer.listeners:
            LOG.debug('No listeners found to update.')
            return
        if amphora is None or amphora.status == consts.DELETED:
            return

        # Check which HAProxy version is on the amp
        haproxy_versions = self._get_haproxy_versions(
            amphora, timeout_dict=timeout_dict)
        # Check if version is supported
        self._populate_amphora_api_version(amphora)

        has_tcp = False
        certs = {}
        listeners_to_update = []
        for listener in loadbalancer.listeners:
            LOG.debug("%s updating listener %s on amphora %s",
                      self.__class__.__name__, listener.id, amphora.id)
            if listener.protocol in consts.LVS_PROTOCOLS:
                # Generate Keepalived LVS configuration from listener object
                config = self.lvs_jinja.build_config(listener=listener)
                self.clients[amphora.api_version].upload_udp_config(
                    amphora, listener.id, config, timeout_dict=timeout_dict)
                self.clients[amphora.api_version].reload_listener(
                    amphora, listener.id, timeout_dict=timeout_dict)
            else:
                has_tcp = True
                obj_id = loadbalancer.id

                try:
                    certs.update({
                        listener.tls_certificate_id:
                        self._process_tls_certificates(
                            listener, amphora, obj_id)['tls_cert']})
                    certs.update({listener.client_ca_tls_certificate_id:
                                  self._process_secret(
                                      listener,
                                      listener.client_ca_tls_certificate_id,
                                      amphora, obj_id)})
                    certs.update({listener.client_crl_container_id:
                                  self._process_secret(
                                      listener,
                                      listener.client_crl_container_id,
                                      amphora, obj_id)})

                    certs.update(self._process_listener_pool_certs(
                        listener, amphora, obj_id))

                    listeners_to_update.append(listener)
                except Exception as e:
                    LOG.exception('Unable to update listener %s due to '
                                  '"%s". Skipping this listener.',
                                  listener.id, str(e))
                    listener_repo = repo.ListenerRepository()
                    with db_api.session().begin() as session:
                        listener_repo.update(session, listener.id,
                                             provisioning_status=consts.ERROR,
                                             operating_status=consts.ERROR)

        if has_tcp:
            if listeners_to_update:
                # Generate HaProxy configuration from listener object
                amp_details = self.clients[amphora.api_version].get_details(
                    amphora)
                config = self.jinja_combo.build_config(
                    host_amphora=amphora, listeners=listeners_to_update,
                    tls_certs=certs,
                    haproxy_versions=haproxy_versions,
                    amp_details=amp_details)
                self.clients[amphora.api_version].upload_config(
                    amphora, loadbalancer.id, config,
                    timeout_dict=timeout_dict)
                self.clients[amphora.api_version].reload_listener(
                    amphora, loadbalancer.id, timeout_dict=timeout_dict)
            else:
                # If we aren't updating any listeners, make sure there are
                # no listeners hanging around. For example if this update
                # was called from a listener delete.
                self.clients[amphora.api_version].delete_listener(
                    amphora, loadbalancer.id)

    def _udp_update(self, listener, vip):
        LOG.debug("Amphora %s keepalivedlvs, updating "
                  "listener %s, vip %s",
                  self.__class__.__name__, listener.protocol_port,
                  vip.ip_address)

        for amp in listener.load_balancer.amphorae:
            if amp.status != consts.DELETED:
                # Generate Keepalived LVS configuration from listener object
                self._populate_amphora_api_version(amp)
                config = self.lvs_jinja.build_config(listener=listener)
                self.clients[amp.api_version].upload_udp_config(
                    amp, listener.id, config)
                self.clients[amp.api_version].reload_listener(
                    amp, listener.id)

    def update(self, loadbalancer):
        for amphora in loadbalancer.amphorae:
            if amphora.status != consts.DELETED:
                self.update_amphora_listeners(loadbalancer, amphora)

    def upload_cert_amp(self, amp, pem):
        LOG.debug("Amphora %s updating cert in REST driver "
                  "with amphora id %s,",
                  self.__class__.__name__, amp.id)
        self._populate_amphora_api_version(amp)
        self.clients[amp.api_version].update_cert_for_rotation(amp, pem)

    def _apply(self, func_name, loadbalancer, amphora=None, *args, **kwargs):
        if amphora is None:
            amphorae = loadbalancer.amphorae
        else:
            amphorae = [amphora]

        for amp in amphorae:
            if amp.status != consts.DELETED:
                self._populate_amphora_api_version(
                    amp, timeout_dict=args[0])
                has_tcp = False
                for listener in loadbalancer.listeners:
                    if listener.protocol in consts.LVS_PROTOCOLS:
                        getattr(self.clients[amp.api_version], func_name)(
                            amp, listener.id, *args)
                    else:
                        has_tcp = True
                if has_tcp:
                    getattr(self.clients[amp.api_version], func_name)(
                        amp, loadbalancer.id, *args)

    def reload(self, loadbalancer, amphora=None, timeout_dict=None):
        self._apply('reload_listener', loadbalancer, amphora, timeout_dict)

    def start(self, loadbalancer, amphora=None, timeout_dict=None):
        self._apply('start_listener', loadbalancer, amphora, timeout_dict)

    def delete(self, listener):
        # Delete any UDP/SCTP listeners the old way (we didn't update the way
        # they are configured)
        loadbalancer = listener.load_balancer
        if listener.protocol in consts.LVS_PROTOCOLS:
            for amp in loadbalancer.amphorae:
                if amp.status != consts.DELETED:
                    self._populate_amphora_api_version(amp)
                    self.clients[amp.api_version].delete_listener(
                        amp, listener.id)
            return

        # In case the listener is not UDP or SCTP, things get more complicated.
        for amp in loadbalancer.amphorae:
            if amp.status != consts.DELETED:
                self._combined_config_delete(amp, listener)

    def _combined_config_delete(self, amphora, listener):
        # Remove the listener from the listener list on the LB before
        # passing the whole thing over to update (so it'll actually delete)
        # In case of amphorae in ACTIVE_STANDBY topology, ensure that we don't
        # remove an already removed listener.
        if listener in listener.load_balancer.listeners:
            listener.load_balancer.listeners.remove(listener)

        # Check if there's any certs that we need to delete
        certs = self._process_tls_certificates(listener)
        certs_to_delete = set()
        if certs['tls_cert']:
            certs_to_delete.add(certs['tls_cert'].id)
        for sni_cert in certs['sni_certs']:
            certs_to_delete.add(sni_cert.id)

        # Delete them (they'll be recreated before the reload if they are
        # needed for other listeners anyway)
        self._populate_amphora_api_version(amphora)
        for cert_id in certs_to_delete:
            self.clients[amphora.api_version].delete_cert_pem(
                amphora, listener.load_balancer.id,
                f'{cert_id}.pem')

        # See how many non-UDP/SCTP listeners we have left
        non_lvs_listener_count = len([
            1 for li in listener.load_balancer.listeners
            if li.protocol not in consts.LVS_PROTOCOLS])
        if non_lvs_listener_count > 0:
            # We have other listeners, so just update is fine.
            # TODO(rm_work): This is a little inefficient since this duplicates
            # a lot of the detection logic that has already been done, but it
            # is probably safer to reuse the existing code-path.
            self.update_amphora_listeners(listener.load_balancer, amphora)
        else:
            # Deleting the last listener, so really do the delete
            self.clients[amphora.api_version].delete_listener(
                amphora, listener.load_balancer.id)

    def get_info(self, amphora, raise_retry_exception=False,
                 timeout_dict=None):
        self._populate_amphora_api_version(
            amphora, raise_retry_exception=raise_retry_exception,
            timeout_dict=timeout_dict)
        return self.clients[amphora.api_version].get_info(
            amphora, raise_retry_exception=raise_retry_exception,
            timeout_dict=timeout_dict)

    def get_diagnostics(self, amphora):
        pass

    def finalize_amphora(self, amphora):
        pass

    def _build_net_info(self, port, amphora, subnet, mtu=None, sriov=False):
        # NOTE(blogan): using the vrrp port here because that
        # is what the allowed address pairs network driver sets
        # this particular port to.  This does expose a bit of
        # tight coupling between the network driver and amphora
        # driver.  We will need to revisit this to try and remove
        # this tight coupling.
        # NOTE (johnsom): I am loading the vrrp_ip into the
        # net_info structure here so that I don't break
        # compatibility with old amphora agent versions.
        host_routes = [{'nexthop': hr[consts.NEXTHOP],
                        'destination': hr[consts.DESTINATION]}
                       for hr in subnet[consts.HOST_ROUTES]]
        net_info = {'subnet_cidr': subnet[consts.CIDR],
                    'gateway': subnet[consts.GATEWAY_IP],
                    'mac_address': port[consts.MAC_ADDRESS],
                    'vrrp_ip': amphora[consts.VRRP_IP],
                    'mtu': mtu or port[consts.NETWORK][consts.MTU],
                    'host_routes': host_routes,
                    'additional_vips': [],
                    'is_sriov': sriov}
        return net_info

    def post_vip_plug(self, amphora, load_balancer, amphorae_network_config,
                      vrrp_port, vip_subnet, additional_vip_data=None):
        if amphora.status != consts.DELETED:
            self._populate_amphora_api_version(amphora)
            port = vrrp_port.to_dict(recurse=True)
            mtu = port[consts.NETWORK][consts.MTU]
            LOG.debug("Post-VIP-Plugging with vrrp_ip %s vrrp_port %s",
                      amphora.vrrp_ip, port[consts.ID])
            sriov = False
            if load_balancer.vip.vnic_type == consts.VNIC_TYPE_DIRECT:
                sriov = True
            net_info = self._build_net_info(
                port, amphora.to_dict(),
                vip_subnet.to_dict(recurse=True), mtu, sriov)
            for add_vip in additional_vip_data:
                add_host_routes = [{'nexthop': hr.nexthop,
                                    'destination': hr.destination}
                                   for hr in add_vip.subnet.host_routes]
                add_net_info = {'subnet_cidr': add_vip.subnet.cidr,
                                'ip_address': add_vip.ip_address,
                                'gateway': add_vip.subnet.gateway_ip,
                                'host_routes': add_host_routes}
                net_info['additional_vips'].append(add_net_info)
            try:
                self.clients[amphora.api_version].plug_vip(
                    amphora, load_balancer.vip.ip_address, net_info)
            except exc.Conflict:
                LOG.warning('VIP with MAC %(mac)s already exists on amphora, '
                            'skipping post_vip_plug',
                            {'mac': port[consts.MAC_ADDRESS]})

    def post_network_plug(self, amphora, port, amphora_network_config):
        fixed_ips = []
        for fixed_ip in port.fixed_ips:
            host_routes = [{'nexthop': hr.nexthop,
                            'destination': hr.destination}
                           for hr in fixed_ip.subnet.host_routes]
            ip = {'ip_address': fixed_ip.ip_address,
                  'subnet_cidr': fixed_ip.subnet.cidr,
                  'host_routes': host_routes,
                  'gateway': fixed_ip.subnet.gateway_ip}
            fixed_ips.append(ip)
        port_info = {'mac_address': port.mac_address,
                     'fixed_ips': fixed_ips,
                     'mtu': port.network.mtu,
                     'is_sriov': False}
        if port.vnic_type == consts.VNIC_TYPE_DIRECT:
            port_info['is_sriov'] = True
        if port.id == amphora.vrrp_port_id:
            # We have to special-case sharing the vrrp port and pass through
            # enough extra information to populate the whole VIP port
            net_info = self._build_net_info(
                port.to_dict(recurse=True), amphora.to_dict(),
                amphora_network_config[consts.VIP_SUBNET],
                port.network.mtu)
            net_info['vip'] = amphora.ha_ip
            port_info['vip_net_info'] = net_info
        try:
            self._populate_amphora_api_version(amphora)
            self.clients[amphora.api_version].plug_network(amphora, port_info)
        except exc.Conflict:
            LOG.warning('Network with MAC %(mac)s already exists on amphora, '
                        'skipping post_network_plug',
                        {'mac': port.mac_address})

    def _process_tls_certificates(self, listener, amphora=None, obj_id=None):
        """Processes TLS data from the listener.

        Converts and uploads PEM data to the Amphora API

        return TLS_CERT and SNI_CERTS
        """
        tls_cert = None
        sni_certs = []
        certs = []
        cert_filename_list = []

        data = cert_parser.load_certificates_data(
            self.cert_manager, listener)
        if data['tls_cert'] is not None:
            tls_cert = data['tls_cert']
            # Note, the first cert is the TLS default cert
            certs.append(tls_cert)
        if data['sni_certs']:
            sni_certs = data['sni_certs']
            certs.extend(sni_certs)

        if amphora and obj_id:
            for cert in certs:
                pem = cert_parser.build_pem(cert)
                md5sum = hashlib.md5(
                    pem, usedforsecurity=False).hexdigest()  # nosec
                name = f'{cert.id}.pem'
                cert_filename_list.append(
                    os.path.join(
                        CONF.haproxy_amphora.base_cert_dir, obj_id, name))
                self._upload_cert(amphora, obj_id, pem, md5sum, name)

            if certs:
                # Build and upload the crt-list file for haproxy
                crt_list = "\n".join(cert_filename_list)
                crt_list = f'{crt_list}\n'.encode()
                md5sum = hashlib.md5(
                    crt_list, usedforsecurity=False).hexdigest()  # nosec
                name = f'{listener.id}.pem'
                self._upload_cert(amphora, obj_id, crt_list, md5sum, name)
        return {'tls_cert': tls_cert, 'sni_certs': sni_certs}

    def _process_secret(self, listener, secret_ref, amphora=None, obj_id=None):
        """Get the secret from the cert manager and upload it to the amp.

        :returns: The filename of the secret in the amp.
        """
        if not secret_ref:
            return None
        context = oslo_context.RequestContext(project_id=listener.project_id)
        secret = self.cert_manager.get_secret(context, secret_ref)
        try:
            secret = secret.encode('utf-8')
        except AttributeError:
            pass
        md5sum = hashlib.md5(
            secret, usedforsecurity=False).hexdigest()  # nosec
        id = hashlib.sha1(secret).hexdigest()  # nosec
        name = f'{id}.pem'

        if amphora and obj_id:
            self._upload_cert(
                amphora, obj_id, pem=secret, md5sum=md5sum, name=name)
        return name

    def _process_listener_pool_certs(self, listener, amphora, obj_id):
        #     {'POOL-ID': {
        #         'client_cert': client_full_filename,
        #         'ca_cert': ca_cert_full_filename,
        #         'crl': crl_full_filename}}
        pool_certs_dict = {}
        for pool in listener.pools:
            if pool.id not in pool_certs_dict:
                pool_certs_dict[pool.id] = self._process_pool_certs(
                    listener, pool, amphora, obj_id)
        for l7policy in listener.l7policies:
            if (l7policy.redirect_pool and
                    l7policy.redirect_pool.id not in pool_certs_dict):
                pool_certs_dict[l7policy.redirect_pool.id] = (
                    self._process_pool_certs(listener, l7policy.redirect_pool,
                                             amphora, obj_id))
        return pool_certs_dict

    def _process_pool_certs(self, listener, pool, amphora, obj_id):
        pool_cert_dict = {}

        # Handle the client cert(s) and key
        if pool.tls_certificate_id:
            data = cert_parser.load_certificates_data(self.cert_manager, pool)
            tls_cert = data['tls_cert']
            pem = cert_parser.build_pem(tls_cert)
            try:
                pem = pem.encode('utf-8')
            except AttributeError:
                pass
            md5sum = hashlib.md5(
                pem, usedforsecurity=False).hexdigest()  # nosec
            name = f'{tls_cert.id}.pem'
            if amphora and obj_id:
                self._upload_cert(amphora, obj_id, pem=pem,
                                  md5sum=md5sum, name=name)
            pool_cert_dict['client_cert'] = os.path.join(
                CONF.haproxy_amphora.base_cert_dir, obj_id, name)
        if pool.ca_tls_certificate_id:
            name = self._process_secret(listener, pool.ca_tls_certificate_id,
                                        amphora, obj_id)
            pool_cert_dict['ca_cert'] = os.path.join(
                CONF.haproxy_amphora.base_cert_dir, obj_id, name)
        if pool.crl_container_id:
            name = self._process_secret(listener, pool.crl_container_id,
                                        amphora, obj_id)
            pool_cert_dict['crl'] = os.path.join(
                CONF.haproxy_amphora.base_cert_dir, obj_id, name)

        return pool_cert_dict

    def _upload_cert(self, amp, listener_id, pem, md5sum, name):
        try:
            if self.clients[amp.api_version].get_cert_md5sum(
                    amp, listener_id, name, ignore=(404,)) == md5sum:
                return
        except exc.NotFound:
            pass

        self.clients[amp.api_version].upload_cert_pem(
            amp, listener_id, name, pem)

    def update_amphora_agent_config(self, amphora, agent_config,
                                    timeout_dict=None):
        """Update the amphora agent configuration file.

        :param amphora: The amphora to update.
        :type amphora: object
        :param agent_config: The new amphora agent configuration.
        :type agent_config: string
        :param timeout_dict: Dictionary of timeout values for calls to the
                             amphora. May contain: req_conn_timeout,
                             req_read_timeout, conn_max_retries,
                             conn_retry_interval
        :returns: None

        Note: This will mutate the amphora agent config and adopt the
              new values.
        """
        try:
            self._populate_amphora_api_version(amphora)
            self.clients[amphora.api_version].update_agent_config(
                amphora, agent_config, timeout_dict=timeout_dict)
        except exc.NotFound as e:
            LOG.debug('Amphora %s does not support the update_agent_config '
                      'API.', amphora.id)
            raise driver_except.AmpDriverNotImplementedError() from e

    def get_interface_from_ip(self, amphora, ip_address, timeout_dict=None):
        """Get the interface name for an IP address.

        :param amphora: The amphora to query.
        :type amphora: octavia.db.models.Amphora
        :param ip_address: The IP address to lookup. (IPv4 or IPv6)
        :type ip_address: string
        :param timeout_dict: Dictionary of timeout values for calls to the
                             amphora. May contain: req_conn_timeout,
                             req_read_timeout, conn_max_retries,
                             conn_retry_interval
        :type timeout_dict: dict
        :returns: the interface name string if found.
        :raises octavia.amphorae.drivers.haproxy.exceptions.NotFound:
                No interface found on the amphora
        :raises TimeOutException: The amphora didn't reply
        """
        self._populate_amphora_api_version(amphora, timeout_dict)
        response_json = self.clients[amphora.api_version].get_interface(
            amphora, ip_address, timeout_dict, log_error=False)
        return response_json.get('interface', None)

    def set_interface_rules(self, amphora: db_models.Amphora,
                            ip_address, rules, timeout_dict=None):
        """Sets interface firewall rules in the amphora

        :param amphora: The amphora to query.
        :param ip_address: The IP address assigned to the interface the rules
                           will be applied on.
        :param rules: The l1st of allow rules to apply.
        """
        try:
            self._populate_amphora_api_version(amphora, timeout_dict)
            self.clients[amphora.api_version].set_interface_rules(
                amphora, ip_address, rules, timeout_dict=timeout_dict)
        except exc.NotFound as e:
            LOG.debug('Amphora %s does not support the set_interface_rules '
                      'API.', amphora.id)
            raise driver_except.AmpDriverNotImplementedError() from e


# Check a custom hostname
class CustomHostNameCheckingAdapter(requests.adapters.HTTPAdapter):
    def cert_verify(self, conn, url, verify, cert):
        conn.assert_hostname = self.uuid
        return super().cert_verify(conn, url, verify, cert)

    def init_poolmanager(self, *pool_args, **pool_kwargs):
        proto = CONF.amphora_agent.agent_tls_protocol.replace('.', '_')
        pool_kwargs['ssl_version'] = getattr(ssl, f"PROTOCOL_{proto}")
        return super().init_poolmanager(*pool_args, **pool_kwargs)


class AmphoraAPIClientBase:
    def __init__(self):
        super().__init__()

        self.get = functools.partial(self.request, 'get')
        self.post = functools.partial(self.request, 'post')
        self.put = functools.partial(self.request, 'put')
        self.delete = functools.partial(self.request, 'delete')
        self.head = functools.partial(self.request, 'head')

        self.session = requests.Session()
        self.session.cert = CONF.haproxy_amphora.client_cert
        self.ssl_adapter = CustomHostNameCheckingAdapter()
        self.session.mount('https://', self.ssl_adapter)

    def _base_url(self, ip, api_version=None):
        if utils.is_ipv6_lla(ip):
            ip = f'[{ip}%{CONF.haproxy_amphora.lb_network_interface}]'
        elif utils.is_ipv6(ip):
            ip = f'[{ip}]'
        if api_version:
            return (f"https://{ip}:{CONF.haproxy_amphora.bind_port}"
                    f"/{api_version}/")
        return f"https://{ip}:{CONF.haproxy_amphora.bind_port}/"

    def request(self, method: str, amp: db_models.Amphora, path: str = '/',
                timeout_dict: Optional[dict] = None,
                retry_404: bool = True, raise_retry_exception: bool = False,
                **kwargs):
        cfg_ha_amp = CONF.haproxy_amphora
        if timeout_dict is None:
            timeout_dict = {}
        req_conn_timeout = timeout_dict.get(
            consts.REQ_CONN_TIMEOUT, cfg_ha_amp.rest_request_conn_timeout)
        req_read_timeout = timeout_dict.get(
            consts.REQ_READ_TIMEOUT, cfg_ha_amp.rest_request_read_timeout)
        conn_max_retries = timeout_dict.get(
            consts.CONN_MAX_RETRIES, cfg_ha_amp.connection_max_retries)
        conn_retry_interval = timeout_dict.get(
            consts.CONN_RETRY_INTERVAL, cfg_ha_amp.connection_retry_interval)

        LOG.debug("request url %s", path)
        _request = getattr(self.session, method.lower())
        _url = self._base_url(amp.lb_network_ip, amp.api_version) + path
        LOG.debug("request url %s", _url)
        reqargs = {
            'verify': CONF.haproxy_amphora.server_ca,
            'url': _url,
            'timeout': (req_conn_timeout, req_read_timeout), }
        reqargs.update(kwargs)
        headers = reqargs.setdefault('headers', {})

        headers['User-Agent'] = (
            f"Octavia HaProxy Rest Client/{amp.api_version} "
            f"(https://wiki.openstack.org/wiki/Octavia)")
        self.ssl_adapter.uuid = amp.id
        exception = None
        # Keep retrying
        for dummy in range(conn_max_retries):
            try:
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore",
                        message="A true SSLContext object is not available"
                    )
                    r = _request(**reqargs)
                LOG.debug('Connected to amphora. Response: %(resp)s',
                          {'resp': r})

                content_type = r.headers.get('content-type', '')
                # Check the 404 to see if it is just that the network in the
                # amphora is not yet up, in which case retry.
                # Otherwise return the response quickly.
                if r.status_code == 404:
                    if not retry_404:
                        raise exc.NotFound()
                    LOG.debug('Got a 404 (content-type: %(content_type)s) -- '
                              'connection data: %(content)s',
                              {'content_type': content_type,
                               'content': r.content})
                    if content_type.find("application/json") == -1:
                        LOG.debug("Amphora agent not ready.")
                        raise requests.ConnectionError
                    try:
                        json_data = r.json().get('details', '')
                        if 'No suitable network interface found' in json_data:
                            LOG.debug("Amphora network interface not found.")
                            raise requests.ConnectionError
                    except json.JSONDecodeError:  # if r.json() fails
                        pass  # TODO(rm_work) Should we do something?
                return r
            except (requests.ConnectionError, requests.Timeout) as e:
                exception = e
                LOG.warning("Could not connect to instance. Retrying.")
                time.sleep(conn_retry_interval)
                if raise_retry_exception:
                    # For taskflow persistence cause attribute should
                    # be serializable to JSON. Pass None, as cause exception
                    # is described in the expection message.
                    raise driver_except.AmpConnectionRetry(
                        exception=str(e)) from None
        LOG.error("Connection retries (currently set to %(max_retries)s) "
                  "exhausted.  The amphora is unavailable. Reason: "
                  "%(exception)s",
                  {'max_retries': conn_max_retries,
                   'exception': exception})
        raise driver_except.TimeOutException()

    def get_api_version(self, amp, timeout_dict=None,
                        raise_retry_exception=False):
        amp.api_version = None
        r = self.get(amp, retry_404=False, timeout_dict=timeout_dict,
                     raise_retry_exception=raise_retry_exception)
        # Handle 404 special as we don't want to log an ERROR on 404
        exc.check_exception(r, (404,))
        if r.status_code == 404:
            raise exc.NotFound()
        return r.json()


class AmphoraAPIClient1_0(AmphoraAPIClientBase):
    def __init__(self):
        super().__init__()

        self.start_listener = functools.partial(self._action,
                                                consts.AMP_ACTION_START)
        self.reload_listener = functools.partial(self._action,
                                                 consts.AMP_ACTION_RELOAD)

        self.start_vrrp = functools.partial(self._vrrp_action,
                                            consts.AMP_ACTION_START)
        self.stop_vrrp = functools.partial(self._vrrp_action,
                                           consts.AMP_ACTION_STOP)
        self.reload_vrrp = functools.partial(self._vrrp_action,
                                             consts.AMP_ACTION_RELOAD)

    def upload_config(self, amp, loadbalancer_id, config, timeout_dict=None):
        r = self.put(
            amp,
            f'loadbalancer/{amp.id}/{loadbalancer_id}/haproxy',
            timeout_dict, data=config)
        return exc.check_exception(r)

    def get_listener_status(self, amp, listener_id):
        r = self.get(
            amp,
            f'listeners/{listener_id}')
        if exc.check_exception(r):
            return r.json()
        return None

    def _action(self, action, amp, object_id, timeout_dict=None):
        r = self.put(
            amp, f'loadbalancer/{object_id}/{action}',
            timeout_dict=timeout_dict)
        return exc.check_exception(r)

    def upload_cert_pem(self, amp, loadbalancer_id, pem_filename, pem_file):
        r = self.put(
            amp,
            f'loadbalancer/{loadbalancer_id}/certificates/{pem_filename}',
            data=pem_file)
        return exc.check_exception(r)

    def get_cert_md5sum(self, amp, loadbalancer_id, pem_filename,
                        ignore=tuple()):
        r = self.get(
            amp,
            f'loadbalancer/{loadbalancer_id}/certificates/{pem_filename}')
        if exc.check_exception(r, ignore):
            return r.json().get("md5sum")
        return None

    def delete_cert_pem(self, amp, loadbalancer_id, pem_filename):
        r = self.delete(
            amp,
            f'loadbalancer/{loadbalancer_id}/certificates/{pem_filename}')
        return exc.check_exception(r, (404,))

    def update_cert_for_rotation(self, amp, pem_file):
        r = self.put(amp, 'certificate', data=pem_file)
        return exc.check_exception(r)

    def delete_listener(self, amp, object_id):
        r = self.delete(
            amp, f'listeners/{object_id}')
        return exc.check_exception(r, (404,))

    def get_info(self, amp, raise_retry_exception=False,
                 timeout_dict=None):
        r = self.get(amp, "info", raise_retry_exception=raise_retry_exception,
                     timeout_dict=timeout_dict)
        if exc.check_exception(r):
            return r.json()
        return None

    def get_details(self, amp):
        r = self.get(amp, "details")
        if exc.check_exception(r):
            return r.json()
        return None

    def get_all_listeners(self, amp):
        r = self.get(amp, "listeners")
        if exc.check_exception(r):
            return r.json()
        return None

    def plug_network(self, amp, port):
        r = self.post(amp, 'plug/network',
                      json=port)
        return exc.check_exception(r)

    def plug_vip(self, amp, vip, net_info):
        r = self.post(amp,
                      f'plug/vip/{vip}',
                      json=net_info)
        return exc.check_exception(r)

    def upload_vrrp_config(self, amp, config):
        r = self.put(amp, 'vrrp/upload', data=config)
        return exc.check_exception(r)

    def _vrrp_action(self, action, amp, timeout_dict=None):
        r = self.put(amp, f'vrrp/{action}',
                     timeout_dict=timeout_dict)
        return exc.check_exception(r)

    def get_interface(self, amp, ip_addr, timeout_dict=None, log_error=True):
        r = self.get(amp, f'interface/{ip_addr}',
                     timeout_dict=timeout_dict)
        return exc.check_exception(r, log_error=log_error).json()

    # The function is used for all LVS-supported protocol listener (UDP, SCTP)
    def upload_udp_config(self, amp, listener_id, config, timeout_dict=None):
        r = self.put(
            amp,
            f'listeners/{amp.id}/{listener_id}/udp_listener', timeout_dict,
            data=config)
        return exc.check_exception(r)

    def update_agent_config(self, amp, agent_config, timeout_dict=None):
        r = self.put(amp, 'config', timeout_dict, data=agent_config)
        return exc.check_exception(r)

    def set_interface_rules(self, amp, ip_address, rules, timeout_dict=None):
        r = self.put(amp, f'interface/{ip_address}/rules', timeout_dict,
                     json=rules)
        return exc.check_exception(r)
